#include <pw/core/axisangle.hpp>
#include <pw/core/matrix_transform.hpp>
#include <pw/core/primitives.hpp>
#include <pw/core/serialize.hpp>
#include <pw/core/vector.hpp>

#include <print>
#include <variant>
#include <vector>

namespace pw {

struct attribute final {
    using mask_type    = std::vector<bool>;
    using int8_type    = std::vector<int8_t>;
    using int32_type   = std::vector<int32_t>;
    using float_type   = std::vector<float>;
    using vector2_type = std::vector<vector2<float>>;
    using vector3_type = std::vector<vector3<float>>;
    using vector4_type = std::vector<vector4<float>>;

    using attribute_data =
        std::variant<std::monostate, mask_type, int8_type, int32_type,
                     float_type, vector2_type, vector3_type, vector4_type>;

    enum attribute_type {
        normals,
        texture_coordinates,
    };

    attribute_type type{};
    attribute_data data{};
};

struct mesh_n {
    primitives geometry{};
    std::vector<attribute> attributes;
};

} // namespace pw

auto main() -> int {

    auto geom = pw::primitives{
        .vertices = {{1.2f, 2.4f, 4.8f},
                     {2.4f, 1.2f, 4.8f},
                     {1.2f, 4.8f, 2.4f}},
        .indices  = {0, 1, 2},
        .topology = pw::primitives::topology_type::triangle_list};

    auto mesh = pw::mesh_n{.geometry = geom};

    mesh.attributes.emplace_back(
        pw::attribute{.type = pw::attribute::normals,
                      .data = std::vector{
                          pw::vector{1.2f, 2.4f, 4.8f}}});

    for (auto i : geom.indices) {
        std::print("({}) ", pw::serialize::to_string(geom.vertices[i]));
    }

    // pw::geometry geo;

    // pw::vector3_array vs = {
    //     { -1, 1, 0 },
    //     { -1,-1, 0 },
    //     {  1,-1, 0 },
    //     {  1, 1, 0 }
    // };

    // pw::geometry::indices_t idx = {
    //     0, 1, 2,
    //     0, 2, 3
    // };

    // geo.set_vertices(vs);
    // geo.set_indices(idx);

    // geo.compute_normals();

    //    amesh.set_vertices(vs);

    //    for (auto v : amesh.vertices()) {
    //        std::cout << pw::serialize::matrix(v.transposed()) << std::endl;
    //    }

    //    auto scale = pw::matrix_transform<float>::scale_matrix({2,2,2});

    //    amesh.transform(scale);

    //    std::cout << "after scale" << std::endl;

    //    for (auto v : amesh.vertices()) {
    //        std::cout << pw::serialize::matrix(v.transposed()) << std::endl;
    //    }

    //    pw::axisangle aa;

    //    aa.axis = pw::vector3({ 0, 0, 1 });
    //    aa.angle = pw::deg_to_rad(90.0f);

    //    auto rot_mat = aa.to_matrix();

    //    amesh.transform(rot_mat);

    //    std::cout << "after rotate" << std::endl;

    //    for (auto v : amesh.vertices()) {
    //        std::cout << pw::serialize::matrix(v.transposed()) << std::endl;
    //    }
}
